/*
 * @OSF_COPYRIGHT@
 * COPYRIGHT NOTICE
 * Copyright (c) 1990, 1991, 1992, 1993, 1996 Open Software Foundation, Inc.
 * ALL RIGHTS RESERVED (DCE).  See the file named COPYRIGHT.DCE in the
 * src directory for the full copyright text.
 */
/*
 * HISTORY
 * $Log: d20fmodify.c,v $
 * Revision 1.1.10.2  1996/02/18  19:38:50  marty
 * 	Update OSF copyright years
 * 	[1996/02/18  18:19:48  marty]
 *
 * Revision 1.1.10.1  1995/12/08  15:11:52  root
 * 	Submit OSF/DCE 1.2.1
 * 	[1995/12/08  14:49:55  root]
 * 
 * Revision 1.1.8.2  1994/03/23  15:03:01  keutel
 * 	March 1994 code drop
 * 	[1994/03/21  16:02:29  keutel]
 * 
 * Revision 1.1.8.1  1994/02/22  15:56:27  marrek
 * 	Code drop for dce1.1.
 * 	[1994/02/08  09:14:59  marrek]
 * 
 * Revision 1.1.6.2  1993/08/10  08:43:23  marrek
 * 	July 1993 code drop.
 * 	[1993/07/30  08:09:58  marrek]
 * 
 * Revision 1.1.4.2  1992/12/31  17:45:36  bbelch
 * 	Embedding copyright notice
 * 	[1992/12/31  12:58:57  bbelch]
 * 
 * Revision 1.1.2.2  1992/06/01  21:31:33  zeliff
 * 	New version of file from GDS drop
 * 	[1992/05/28  19:08:09  zeliff]
 * 
 * $EndLog$
 */
#if !defined(lint) && defined(GDS_RCS_ID)
static char rcsid[] = "@(#)$RCSfile: d20fmodify.c,v $ $Revision: 1.1.10.2 $ $Date: 1996/02/18 19:38:50 $";
#endif

/****************************************************************************
*                                                                           *
*         COPYRIGHT  (C)  SIEMENS NIXDORF INFORMATIONSSYSTEME AG 1991       *
*                                ALL RIGHTS RESERVED                        *
*                                                                           *
****************************************************************************/

/*daton	***************************************************************/
/*								      */
/* TYPE         : MODUL                                               */
/*								      */
/* NAME         : d20fmodify.c                                        */
/*								      */
/* AUTHOR       : R. Horn, D AP 11                                    */
/* DATE         : 10.05.88                                            */
/*								      */
/* COMPONENT    : DS                                                  */
/*								      */
/* DOC.-NR.     : Directory-Design-Specification                      */
/*			.					      */
/*			.					      */
/*								      */
/* PRD#/VERS.   :                                                     */
/*								      */
/* DESCRIPTION  :                                                     */
/*								      */
/* SYSTEM DEPENDENCIES: COMMON                                        */
/*								      */
/* HISTORY      :                                                     */
/*								      */
/* Vers.Nr. |  Date   |  Updates                       | KZ | CR# FM# */
/*          |         |                                |    |         */
/*datoff **************************************************************/


/*exon ***********************************************************************/
/*                                                                           */
/*  FUNCTION:  d20_fill_modify ()                                            */
/*                                                                           */
/*  SHORT DESCRIPTION:                                                       */
/*      The function converts one or more attributes, which will be modified */
/*      and which are given in a format used by the  administration, into a  */
/*      'queue'-format used at the IAPL-interface of the directory system V2.*/
/*      The caller is able to control the way in which the conversion is     */
/*      done by a special parameter.                                         */
/*                                                                           */
/*  INPUT-PARAMETER:                                                         */
/*      opmode      = Operation mode:                                        */
/*                    The operation mode is controlled by a 'bit field'-     */
/*                    mask, where a specific mask can be generated by ORing  */
/*                    the following 'bit field'-values:                      */
/*                    D20_ASTART  = initialize attribute output queue.       */
/*                    D20_AINSERT = insert given attribute informations into */
/*                                  attribute output queue.                  */
/*                    D20_AEND    = terminate attribute output queue.        */
/*                                                                           */
/*      attr_ent    = Pointer to an array containing the informations about  */
/*                    the convertible attributes (attribute type, old resp.  */
/*                    new attribute value, length of the old resp. new       */
/*                    attribute value).                                      */
/*                    The following modification modes are allowed:          */
/*                       D2_REP_AV = replace attribute value.                */
/*                       D2_ADD_AV = add attribute value.                    */
/*                       D2_DEL_AV = delete attribute value.                 */
/*                       D2_DELATT = delete attribute                        */
/*                       D2_ADDATT = add attribute                           */
/*                    Note: Two continuous attribute entries correspond al-  */
/*                    ways to a convertible attribute, where the first con-  */
/*                    tains information about the old attribute value and    */
/*                    the second contains information about the new attri-   */
/*                    bute value.                                            */
/*                    In case of D2_DELATT and D2_ADDATT only one attribute  */
/*                    entry is necessary.                                    */
/*                                                                           */
/*      no_attr     = No. of convertible attributes.                         */
/*                                                                           */
/*  OUTPUT-PARAMETER:                                                        */
/*      info        = Converted 'modify' attribute(s).                       */
/*                                                                           */
/*  RETURN-VALUE:                                                            */
/*                    None                                                   */
/*                                                                           */
/* ------------------------------------------------------------------------- */
/*  AUTHOR: R.Horn                                        DATE: 4.3.88       */
/*exoff **********************************************************************/

#include <gds.h>
#include <stdio.h>

#include <dce/d2dir.h>
#include <gdsext.h>
#include <d2adm.h>
#include <d2info.h>
#include <d2spec_msk.h>
#include <d20proto.h>

/* --------------------   declaration of funktions ------------------------- */

/* -------------------- declaration of global data ------------------------- */

extern Ds_v2_info	info ;	/* common IAPL-interface parameter block */

extern D2_a_value       *d20_avptr ;
extern Octet_string     d20_valptr ;

extern int             d20_idx;
extern D2_ttxid        array_ttxid[];
extern D2_T61_seq      array_t61seq[];
extern D2_str          array_str[][D20_MAX_NB_ITEM];
extern D2_faxnumber    array_fax[];
extern D2_telex_number array_telex[];

/* ------------------------------------------------------------------------- */

 void d20_fill_modify (
   signed16     optype,
   Ds_at_entry *attr_ent,
   signed16     no_attr)
{
  static D2_a_modinf      *d20_miptr ;
  register signed16  i;
  unsigned16  rep;
  Bool	copy_val_flag = TRUE;

if ((optype & D20_ASTART) == D20_ASTART) 
	{
	/* initialize modify attribute entry queue */
	d20_miptr = info.modinfo ;
	d20_avptr = info.avalinfo ;
	d20_valptr = info.avalues ;
	d20_idx = -1;
	}

if ((optype & D20_AINSERT) == D20_AINSERT) 
	{
	d20_idx++;

	/* insert attribute information into modify attribute entry queue */
	for (i = 0; i < no_attr; i++, d20_miptr++, attr_ent++) 
		{
		if (attr_ent->mod_type == D2_ADD_AV)
			attr_ent++;

		d20_miptr->d2_m_next = d20_miptr + 1 ;

		d20_miptr->d2_m_mod = attr_ent->mod_type ;
		d20_miptr->d2_m_attr.d2_a_next = (D2_a_info *)NULL;
		d20_miptr->d2_m_attr.d2_a_type.d2_type =
		attr_ent->attr_id.d2_type ;
		d20_miptr->d2_m_attr.d2_a_type.d2_typ_len =
		attr_ent->attr_id.d2_typ_len ;

	        if (attr_ent->mod_type != D2_DELATT)
			{
			d20_miptr->d2_m_attr.d2_a_no_val = 1;
			d20_miptr->d2_m_attr.d2_a_val = d20_avptr ;
			rep = d20_avptr->d2_a_rep = 
					attr_ent->a_val.d2_a_rep;

			if (rep != D2_T61_LIST  && rep != D2_T61_PR_LIST 
			    && rep != D2_FAXNR_STX 
			    && rep != D2_TLXNR_STX && 
			    rep != D2_TTXID_STX && rep != D2_GUIDE && 
			    rep != D2_INTEGER && rep != D2_BOOLEAN && 
			    rep != D2_MHS_ORADR_STX && 
			    rep != D2_MHS_ORNAME_STX && 
			    rep != D2_MHS_DLSP_STX)
				d20_avptr->d2_a_v_len = 
					attr_ent->a_val.d2_a_v_len ;
			else
				d20_avptr->d2_a_v_len = 0;

			if (attr_ent->mod_type == D2_ADDATT)
				copy_val_flag = FALSE;

			if (copy_val_flag == TRUE)
				{
				d20_copy_val(attr_ent);
				d20_idx++;
				}

			d20_avptr++;

			if (attr_ent->mod_type == D2_REP_AV)
				{
				if (attr_ent->a_val.d2_a_rep == D2_INT_LIST)
					{
		        		d20_miptr->d2_m_mod = D2_DELATT;
					d20_miptr->d2_m_attr.d2_a_no_val = 0;

					d20_miptr++;
					
					d20_miptr->d2_m_next = d20_miptr + 1 ;

					d20_miptr->d2_m_mod = D2_ADDATT;
					d20_miptr->d2_m_attr.d2_a_next = 
							(D2_a_info *)NULL;
					d20_miptr->d2_m_attr.d2_a_type.d2_type =
					attr_ent->attr_id.d2_type ;
					d20_miptr->d2_m_attr.d2_a_type.
					d2_typ_len = attr_ent->attr_id.
								d2_typ_len ;

					attr_ent++;

					d20_miptr->d2_m_attr.d2_a_no_val = 1;
					d20_miptr->d2_m_attr.d2_a_val = 
								d20_avptr;
					d20_avptr->d2_a_rep = 
						attr_ent->a_val.d2_a_rep;
					d20_avptr->d2_a_v_len =
						attr_ent->a_val.d2_a_v_len ;

					d20_copy_val(attr_ent);
					d20_avptr++;
					}
				else
					{
					d20_miptr->d2_m_attr.d2_a_no_val = 2;

					attr_ent++;

					rep = d20_avptr->d2_a_rep = 
						attr_ent->a_val.d2_a_rep;
					if (rep != D2_T61_LIST  && 
					    rep != D2_T61_PR_LIST && 
					    rep != D2_FAXNR_STX && 
					    rep != D2_TLXNR_STX && 
					    rep != D2_TTXID_STX && 
					    rep != D2_GUIDE && 
					    rep != D2_INTEGER && 
					    rep != D2_BOOLEAN)
						d20_avptr->d2_a_v_len =
						attr_ent->a_val.d2_a_v_len ;
					else
						d20_avptr->d2_a_v_len = 0;

					d20_copy_val(attr_ent);
					d20_avptr++;
					}
				}
			}
		else
			d20_miptr->d2_m_attr.d2_a_no_val = 0;
		}
	}

if ((optype & D20_AEND) == D20_AEND)
	/* terminate modify attribute entry list */
	(d20_miptr - 1)->d2_m_next = (D2_a_modinf *) NULL ;
}

